/*
* Fabric3
* Copyright (c) 2009 Metaform Systems
*
* Fabric3 is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as
* published by the Free Software Foundation, either version 3 of
* the License, or (at your option) any later version, with the
* following exception:
*
* Linking this software statically or dynamically with other
* modules is making a combined work based on this software.
* Thus, the terms and conditions of the GNU General Public
* License cover the whole combination.
*
* As a special exception, the copyright holders of this software
* give you permission to link this software with independent
* modules to produce an executable, regardless of the license
* terms of these independent modules, and to copy and distribute
* the resulting executable under terms of your choice, provided
* that you also meet, for each linked independent module, the
* terms and conditions of the license of that module. An
* independent module is a module which is not derived from or
* based on this software. If you modify this software, you may
* extend this exception to your version of the software, but
* you are not obligated to do so. If you do not wish to do so,
* delete this exception statement from your version.
*
* Fabric3 is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty
* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU General Public License for more details.
*
* You should have received a copy of the
* GNU General Public License along with Fabric3.
* If not, see <http://www.gnu.org/licenses/>.
*/
package org.fabric3.contribution.manifest;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import org.osoa.sca.annotations.Destroy;
import org.osoa.sca.annotations.EagerInit;
import org.osoa.sca.annotations.Init;
import org.osoa.sca.annotations.Reference;

import org.fabric3.host.Constants;
import org.fabric3.host.Namespaces;
import org.fabric3.host.RuntimeMode;
import org.fabric3.host.contribution.Deployable;
import org.fabric3.spi.contribution.ContributionManifest;
import org.fabric3.spi.contribution.Export;
import org.fabric3.spi.contribution.Import;
import org.fabric3.spi.introspection.IntrospectionContext;
import org.fabric3.spi.introspection.xml.InvalidQNamePrefix;
import org.fabric3.spi.introspection.xml.InvalidValue;
import org.fabric3.spi.introspection.xml.LoaderRegistry;
import org.fabric3.spi.introspection.xml.TypeLoader;
import org.fabric3.spi.introspection.xml.UnrecognizedAttribute;
import org.fabric3.spi.introspection.xml.UnrecognizedElement;
import org.fabric3.spi.introspection.xml.UnrecognizedElementException;

import static javax.xml.stream.XMLStreamConstants.START_ELEMENT;
import static org.oasisopen.sca.Constants.SCA_NS;

/**
 * Loads a contribution manifest from a contribution element
 *
 * @version $Rev: 9343 $ $Date: 2010-08-15 07:46:49 +0000 (Sun, 15 Aug 2010) $
 */
@EagerInit
public class ContributionElementLoader implements TypeLoader<ContributionManifest> {
    private static final QName CONTRIBUTION = new QName(SCA_NS, "contribution");
    private static final QName DEPLOYABLE = new QName(SCA_NS, "deployable");

    private final LoaderRegistry registry;

    public ContributionElementLoader(@Reference LoaderRegistry registry) {
        this.registry = registry;
    }

    @Init
    public void start() {
        registry.registerLoader(CONTRIBUTION, this);
    }

    @Destroy
    public void stop() {
        registry.unregisterLoader(CONTRIBUTION);
    }


    public ContributionManifest load(XMLStreamReader reader, IntrospectionContext context) throws XMLStreamException {
        ContributionManifest manifest = new ContributionManifest();
        QName element = reader.getName();
        if (!CONTRIBUTION.equals(element)) {
            throw new AssertionError("Loader not positioned on the <contribution> element: " + element);
        }
        validateContributionAttributes(reader, context);
        boolean extension = Boolean.valueOf(reader.getAttributeValue(Namespaces.CORE, "extension"));
        manifest.setExtension(extension);

        String description = reader.getAttributeValue(Namespaces.CORE, "description");
        manifest.setDescription(description);

        parseCapabilities(reader, manifest);
        while (true) {
            int event = reader.next();
            switch (event) {
            case START_ELEMENT:
                element = reader.getName();
                if (DEPLOYABLE.equals(element)) {
                    validateDeployableAttributes(reader, context);
                    String name = reader.getAttributeValue(null, "composite");
                    if (name == null) {
                        MissingManifestAttribute failure =
                                new MissingManifestAttribute("Composite attribute must be specified", reader);
                        context.addError(failure);
                        return null;
                    }
                    QName qName;
                    // read qname but only set namespace if it is explicitly declared
                    int index = name.indexOf(':');
                    if (index != -1) {
                        String prefix = name.substring(0, index);
                        String localPart = name.substring(index + 1);
                        String ns = reader.getNamespaceContext().getNamespaceURI(prefix);
                        if (ns == null) {
                            URI uri = context.getContributionUri();
                            context.addError(new InvalidQNamePrefix("The prefix " + prefix + " specified in the contribution manifest file for "
                                    + uri + " is invalid", reader));
                            return null;
                        }
                        qName = new QName(ns, localPart, prefix);
                    } else {
                        qName = new QName(null, name);
                    }
                    List<RuntimeMode> runtimeModes = parseRuntimeModes(reader, context);
                    Deployable deployable = new Deployable(qName, Constants.COMPOSITE_TYPE, runtimeModes);
                    manifest.addDeployable(deployable);
                } else {
                    Object o;
                    try {
                        o = registry.load(reader, Object.class, context);
                    } catch (UnrecognizedElementException e) {
                        UnrecognizedElement failure = new UnrecognizedElement(reader);
                        context.addError(failure);
                        return null;
                    }
                    if (o instanceof Export) {
                        manifest.addExport((Export) o);
                    } else if (o instanceof Import) {
                        manifest.addImport((Import) o);
                    } else if (o instanceof ExtendsDeclaration) {
                        ExtendsDeclaration declaration = (ExtendsDeclaration) o;
                        manifest.addExtend(declaration.getName());
                    } else if (o instanceof ProvidesDeclaration) {
                        ProvidesDeclaration declaration = (ProvidesDeclaration) o;
                        manifest.addExtensionPoint(declaration.getName());
                    } else if (o != null) {
                        UnrecognizedElement failure = new UnrecognizedElement(reader);
                        context.addError(failure);
                        return null;
                    }
                }
                break;
            case XMLStreamConstants.END_ELEMENT:
                if (CONTRIBUTION.equals(reader.getName())) {
                    return manifest;
                }
                break;
            }
        }
    }

    private void parseCapabilities(XMLStreamReader reader, ContributionManifest manifest) {

        String requiresAttr = reader.getAttributeValue(Namespaces.CORE, "required-capabilities");
        if (requiresAttr != null) {
            String[] requires = requiresAttr.trim().split(" ");
            for (String require : requires) {
                manifest.addRequiredCapability(require);
            }
        }

        String providesAttr = reader.getAttributeValue(Namespaces.CORE, "capabilities");
        if (providesAttr != null) {
            String[] provides = providesAttr.trim().split(" ");
            for (String provide : provides) {
                manifest.addProvidedCapability(provide);
            }
        }
    }

    private List<RuntimeMode> parseRuntimeModes(XMLStreamReader reader, IntrospectionContext context) {
        String modeAttr = reader.getAttributeValue(null, "modes");
        List<RuntimeMode> runtimeModes;
        if (modeAttr == null) {
            runtimeModes = Deployable.DEFAULT_MODES;
        } else {
            String[] modes = modeAttr.trim().split(" ");
            runtimeModes = new ArrayList<RuntimeMode>();
            for (String mode : modes) {
                if ("controller".equals(mode)) {
                    runtimeModes.add(RuntimeMode.CONTROLLER);
                } else if ("participant".equals(mode)) {
                    runtimeModes.add(RuntimeMode.PARTICIPANT);
                } else if ("vm".equals(mode)) {
                    runtimeModes.add(RuntimeMode.VM);
                } else {
                    runtimeModes = Deployable.DEFAULT_MODES;
                    InvalidValue error = new InvalidValue("Invalid mode attribute: " + modeAttr, reader);
                    context.addError(error);
                    break;
                }
            }
        }
        return runtimeModes;
    }

    private void validateContributionAttributes(XMLStreamReader reader, IntrospectionContext context) {
        for (int i = 0; i < reader.getAttributeCount(); i++) {
            String name = reader.getAttributeLocalName(i);
            if (!"extension".equals(name) && !"description".equals(name) && !"capabilities".equals(name) && !"required-capabilities".equals(name)) {
                context.addError(new UnrecognizedAttribute(name, reader));
            }
        }
    }

    private void validateDeployableAttributes(XMLStreamReader reader, IntrospectionContext context) {
        for (int i = 0; i < reader.getAttributeCount(); i++) {
            String name = reader.getAttributeLocalName(i);
            if (!"composite".equals(name) && !"modes".equals(name)) {
                context.addError(new UnrecognizedAttribute(name, reader));
            }
        }
    }


}
